@@ -9,7 +9,7 @@ use strict;
 use warnings;
 
 package Net::SIP::StatelessProxy;
-use fields qw( dispatcher rewrite_contact nathelper force_rewrite );
+use fields qw( dispatcher rewrite_contact nathelper force_rewrite respcode );
 
 use Net::SIP::Util ':all';
 use Digest::MD5 qw(md5);
@@ -50,6 +50,7 @@ sub new {
     };
     $self->{nathelper} = delete $args{nathelper};
     $self->{force_rewrite} = delete $args{force_rewrite};
+    $self->{respcode} = [ {},{} ];
 
     return $self;
 }
@@ -641,6 +642,11 @@ sub do_nat {
 	: ( undef,$packet )
 	;
     my $method = $request ? $request->method : '';
+    my $track_resp_code;
+    if ($response and $response->method eq 'INVITE') {
+	my $code = $response->code;
+	$track_resp_code = $code if $code>=400;
+    }
 
     # NAT for anything with SDP body
     # activation and close of session will be done on ACK|CANCEL|BYE
@@ -649,7 +655,7 @@ sub do_nat {
 	or $method eq 'CANCEL'
 	or $method eq 'BYE' ) {
 	DEBUG( 100, "no NAT because no SDP body and method is $method" );
-	return;
+	return if ! $track_resp_code;
     }
 
 
@@ -693,6 +699,22 @@ sub do_nat {
 	or return [ 0,'no CSEQ in packet' ];
     my $callid = $packet->callid;
 
+    if ($track_resp_code) {
+	my $rc = $self->{respcode}[0];
+	if (keys(%$rc)>5000) {
+	    # expire entries
+	    $self->{respcode}[1] = $rc;
+	    $rc = $self->{respcode}[0] = {};
+	}
+	$rc->{$callid,$cseq,$idfrom,$idto} = $track_resp_code;
+
+	# No NAT to do, we just needed to track the response code
+	# The session should be closed though since it will not be completed
+	DEBUG( 50,"close session $callid|$cseq because of ".( $request ? $method : $response->code." $method"));
+	$nathelper->close_session( $callid,$cseq,$idfrom,$idto );
+	return;
+    }
+
     # CANCEL|BYE will be handled first to close session
     # no NAT will be done, even if the packet contains SDP (which makes no sense)
     if ( $method eq 'CANCEL' ) {
@@ -710,6 +732,7 @@ sub do_nat {
     if ( $body ) {
 	DEBUG( 100,"need to NAT SDP body: ".$body->as_string );
 
+	DEBUG( 50,"allocate sockets $callid|$cseq because of SDP body in ".($request ? $method : $response->code));
 	my $new_media = $nathelper->allocate_sockets(
 	    $callid,$cseq,$idfrom,$idto,$side,$outgoing_leg->laddr(0),
 	    scalar( $body->get_media) );
@@ -727,14 +750,20 @@ sub do_nat {
     # In a lot of cases this will be too early, because I only have one
     # site, but only in the case of ACK an incomplete session is invalid.
 
-    if ( ! $nathelper->activate_session( $callid,$cseq,$idfrom,$idto ) ) {
-	if ( $method eq 'ACK' ) {
+    if ( $method eq 'ACK' ) {
+	my $code = $self->{respcode}[0]{$callid,$cseq,$idfrom,$idto}
+	    || $self->{respcode}[1]{$callid,$cseq,$idfrom,$idto}
+	    || -1;
+	if ($code > 400) {
+	    # ACK to response with error code, should be closed already
+	    DEBUG( 100, "session $callid|$cseq $idfrom -> ACK to failure response" );
+	} elsif (! $nathelper->activate_session($callid,$cseq,$idfrom,$idto)) {
 	    DEBUG( 50,"session $callid|$cseq $idfrom -> $idto still incomplete in ACK" );
 	    return [ 0,'incomplete session in ACK' ]
-	} else {
-	    # ignore problem, session not yet complete
-	    DEBUG( 100, "session $callid|$cseq $idfrom -> $idto not yet complete" );
 	}
+    } elsif (! $nathelper->activate_session($callid,$cseq,$idfrom,$idto)) {
+	# ignore problem, session not yet complete
+	DEBUG( 100, "session $callid|$cseq $idfrom -> $idto not yet complete" );
     } else {
 	DEBUG( 50,"activated session $callid|$cseq $idfrom -> $idto" )
     }
